# SPDX-License-Identifier: GPL-2.0
# intel-pt-events.py: Print Intel PT Events including Power Events and PTWRITE
# Copyright (c) 2017-2021, Intel Corporation.
#
# This program is free software; you can redistribute it and/or modify it
# under the terms and conditions of the GNU General Public License,
# version 2, as published by the Free Software Foundation.
#
# This program is distributed in the hope it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
# more details.

from __future__ import division, print_function

import io
import os
import sys
import struct
import argparse
import contextlib

from libxed import LibXED
from ctypes import create_string_buffer, addressof

sys.path.append(os.environ['PERF_EXEC_PATH'] + \
	'/scripts/python/Perf-Trace-Util/lib/Perf/Trace')

from perf_trace_context import perf_set_itrace_options, \
	perf_sample_insn, perf_sample_srccode

try:
	broken_pipe_exception = BrokenPipeError
except:
	broken_pipe_exception = IOError

glb_switch_str		= {}
glb_insn		= False
glb_disassembler	= None
glb_src			= False
glb_source_file_name	= None
glb_line_number		= None
glb_dso			= None
glb_stash_dict		= {}
glb_output		= None
glb_output_pos		= 0
glb_cpu			= -1
glb_time		= 0

def get_optional_null(perf_dict, field):
	if field in perf_dict:
		return perf_dict[field]
	return ""

def get_optional_zero(perf_dict, field):
	if field in perf_dict:
		return perf_dict[field]
	return 0

def get_optional_bytes(perf_dict, field):
	if field in perf_dict:
		return perf_dict[field]
	return bytes()

def get_optional(perf_dict, field):
	if field in perf_dict:
		return perf_dict[field]
	return "[unknown]"

def get_offset(perf_dict, field):
	if field in perf_dict:
		return "+%#x" % perf_dict[field]
	return ""

def trace_begin():
	ap = argparse.ArgumentParser(usage = "", add_help = False)
	ap.add_argument("--insn-trace", action='store_true')
	ap.add_argument("--src-trace", action='store_true')
	ap.add_argument("--all-switch-events", action='store_true')
	ap.add_argument("--interleave", type=int, nargs='?', const=4, default=0)
	global glb_args
	global glb_insn
	global glb_src
	glb_args = ap.parse_args()
	if glb_args.insn_trace:
		print("Intel PT Instruction Trace")
		itrace = "i0nsepwxI"
		glb_insn = True
	elif glb_args.src_trace:
		print("Intel PT Source Trace")
		itrace = "i0nsepwxI"
		glb_insn = True
		glb_src = True
	else:
		print("Intel PT Branch Trace, Power Events, Event Trace and PTWRITE")
		itrace = "bepwxI"
	global glb_disassembler
	try:
		glb_disassembler = LibXED()
	except:
		glb_disassembler = None
	perf_set_itrace_options(perf_script_context, itrace)

def trace_end():
	if glb_args.interleave:
		flush_stashed_output()
	print("End")

def trace_unhandled(event_name, context, event_fields_dict):
		print(' '.join(['%s=%s'%(k,str(v))for k,v in sorted(event_fields_dict.items())]))

def stash_output():
	global glb_stash_dict
	global glb_output_pos
	output_str = glb_output.getvalue()[glb_output_pos:]
	n = len(output_str)
	if n:
		glb_output_pos += n
		if glb_cpu not in glb_stash_dict:
			glb_stash_dict[glb_cpu] = []
		glb_stash_dict[glb_cpu].append(output_str)

def flush_stashed_output():
	global glb_stash_dict
	while glb_stash_dict:
		cpus = list(glb_stash_dict.keys())
		# Output at most glb_args.interleave output strings per cpu
		for cpu in cpus:
			items = glb_stash_dict[cpu]
			countdown = glb_args.interleave
			while len(items) and countdown:
				sys.stdout.write(items[0])
				del items[0]
				countdown -= 1
			if not items:
				del glb_stash_dict[cpu]

def print_ptwrite(raw_buf):
	data = struct.unpack_from("<IQ", raw_buf)
	flags = data[0]
	payload = data[1]
	exact_ip = flags & 1
	try:
		s = payload.to_bytes(8, "little").decode("ascii").rstrip("\x00")
		if not s.isprintable():
			s = ""
	except:
		s = ""
	print("IP: %u payload: %#x" % (exact_ip, payload), s, end=' ')

def print_cbr(raw_buf):
	data = struct.unpack_from("<BBBBII", raw_buf)
	cbr = data[0]
	f = (data[4] + 500) / 1000
	p = ((cbr * 1000 / data[2]) + 5) / 10
	print("%3u  freq: %4u MHz  (%3u%%)" % (cbr, f, p), end=' ')

def print_mwait(raw_buf):
	data = struct.unpack_from("<IQ", raw_buf)
	payload = data[1]
	hints = payload & 0xff
	extensions = (payload >> 32) & 0x3
	print("hints: %#x extensions: %#x" % (hints, extensions), end=' ')

def print_pwre(raw_buf):
	data = struct.unpack_from("<IQ", raw_buf)
	payload = data[1]
	hw = (payload >> 7) & 1
	cstate = (payload >> 12) & 0xf
	subcstate = (payload >> 8) & 0xf
	print("hw: %u cstate: %u sub-cstate: %u" % (hw, cstate, subcstate),
		end=' ')

def print_exstop(raw_buf):
	data = struct.unpack_from("<I", raw_buf)
	flags = data[0]
	exact_ip = flags & 1
	print("IP: %u" % (exact_ip), end=' ')

def print_pwrx(raw_buf):
	data = struct.unpack_from("<IQ", raw_buf)
	payload = data[1]
	deepest_cstate = payload & 0xf
	last_cstate = (payload >> 4) & 0xf
	wake_reason = (payload >> 8) & 0xf
	print("deepest cstate: %u last cstate: %u wake reason: %#x" %
		(deepest_cstate, last_cstate, wake_reason), end=' ')

def print_psb(raw_buf):
	data = struct.unpack_from("<IQ", raw_buf)
	offset = data[1]
	print("offset: %#x" % (offset), end=' ')

glb_cfe = ["", "INTR", "IRET", "SMI", "RSM", "SIPI", "INIT", "VMENTRY", "VMEXIT",
		"VMEXIT_INTR", "SHUTDOWN", "", "UINT", "UIRET"] + [""] * 18
glb_evd = ["", "PFA", "VMXQ", "VMXR"] + [""] * 60

def print_evt(raw_buf):
	data = struct.unpack_from("<BBH", raw_buf)
	typ = data[0] & 0x1f
	ip_flag = (data[0] & 0x80) >> 7
	vector = data[1]
	evd_cnt = data[2]
	s = glb_cfe[typ]
	if s:
		print(" cfe: %s IP: %u vector: %u" % (s, ip_flag, vector), end=' ')
	else:
		print(" cfe: %u IP: %u vector: %u" % (typ, ip_flag, vector), end=' ')
	pos = 4
	for i in range(evd_cnt):
		data = struct.unpack_from("<QQ", raw_buf)
		et = data[0] & 0x3f
		s = glb_evd[et]
		if s:
			print("%s: %#x" % (s, data[1]), end=' ')
		else:
			print("EVD_%u: %#x" % (et, data[1]), end=' ')

def print_iflag(raw_buf):
	data = struct.unpack_from("<IQ", raw_buf)
	iflag = data[0] & 1
	old_iflag = iflag ^ 1
	via_branch = data[0] & 2
	branch_ip = data[1]
	if via_branch:
		s = "via"
	else:
		s = "non"
	print("IFLAG: %u->%u %s branch" % (old_iflag, iflag, s), end=' ')

def common_start_str(comm, sample):
	ts = sample["time"]
	cpu = sample["cpu"]
	pid = sample["pid"]
	tid = sample["tid"]
	if "machine_pid" in sample:
		machine_pid = sample["machine_pid"]
		vcpu = sample["vcpu"]
		return "VM:%5d VCPU:%03d %16s %5u/%-5u [%03u] %9u.%09u  " % (machine_pid, vcpu, comm, pid, tid, cpu, ts / 1000000000, ts %1000000000)
	else:
		return "%16s %5u/%-5u [%03u] %9u.%09u  " % (comm, pid, tid, cpu, ts / 1000000000, ts %1000000000)

def print_common_start(comm, sample, name):
	flags_disp = get_optional_null(sample, "flags_disp")
	# Unused fields:
	# period      = sample["period"]
	# phys_addr   = sample["phys_addr"]
	# weight      = sample["weight"]
	# transaction = sample["transaction"]
	# cpumode     = get_optional_zero(sample, "cpumode")
	print(common_start_str(comm, sample) + "%8s  %21s" % (name, flags_disp), end=' ')

def print_instructions_start(comm, sample):
	if "x" in get_optional_null(sample, "flags"):
		print(common_start_str(comm, sample) + "x", end=' ')
	else:
		print(common_start_str(comm, sample), end='  ')

def disassem(insn, ip):
	inst = glb_disassembler.Instruction()
	glb_disassembler.SetMode(inst, 0) # Assume 64-bit
	buf = create_string_buffer(64)
	buf.value = insn
	return glb_disassembler.DisassembleOne(inst, addressof(buf), len(insn), ip)

def print_common_ip(param_dict, sample, symbol, dso):
	ip   = sample["ip"]
	offs = get_offset(param_dict, "symoff")
	if "cyc_cnt" in sample:
		cyc_cnt = sample["cyc_cnt"]
		insn_cnt = get_optional_zero(sample, "insn_cnt")
		ipc_str = "  IPC: %#.2f (%u/%u)" % (insn_cnt / cyc_cnt, insn_cnt, cyc_cnt)
	else:
		ipc_str = ""
	if glb_insn and glb_disassembler is not None:
		insn = perf_sample_insn(perf_script_context)
		if insn and len(insn):
			cnt, text = disassem(insn, ip)
			byte_str = ("%x" % ip).rjust(16)
			if sys.version_info.major >= 3:
				for k in range(cnt):
					byte_str += " %02x" % insn[k]
			else:
				for k in xrange(cnt):
					byte_str += " %02x" % ord(insn[k])
			print("%-40s  %-30s" % (byte_str, text), end=' ')
		print("%s%s (%s)" % (symbol, offs, dso), end=' ')
	else:
		print("%16x %s%s (%s)" % (ip, symbol, offs, dso), end=' ')
	if "addr_correlates_sym" in sample:
		addr   = sample["addr"]
		dso    = get_optional(sample, "addr_dso")
		symbol = get_optional(sample, "addr_symbol")
		offs   = get_offset(sample, "addr_symoff")
		print("=> %x %s%s (%s)%s" % (addr, symbol, offs, dso, ipc_str))
	else:
		print(ipc_str)

def print_srccode(comm, param_dict, sample, symbol, dso, with_insn):
	ip = sample["ip"]
	if symbol == "[unknown]":
		start_str = common_start_str(comm, sample) + ("%x" % ip).rjust(16).ljust(40)
	else:
		offs = get_offset(param_dict, "symoff")
		start_str = common_start_str(comm, sample) + (symbol + offs).ljust(40)

	if with_insn and glb_insn and glb_disassembler is not None:
		insn = perf_sample_insn(perf_script_context)
		if insn and len(insn):
			cnt, text = disassem(insn, ip)
		start_str += text.ljust(30)

	global glb_source_file_name
	global glb_line_number
	global glb_dso

	source_file_name, line_number, source_line = perf_sample_srccode(perf_script_context)
	if source_file_name:
		if glb_line_number == line_number and glb_source_file_name == source_file_name:
			src_str = ""
		else:
			if len(source_file_name) > 40:
				src_file = ("..." + source_file_name[-37:]) + " "
			else:
				src_file = source_file_name.ljust(41)
			if source_line is None:
				src_str = src_file + str(line_number).rjust(4) + " <source not found>"
			else:
				src_str = src_file + str(line_number).rjust(4) + " " + source_line
		glb_dso = None
	elif dso == glb_dso:
		src_str = ""
	else:
		src_str = dso
		glb_dso = dso

	glb_line_number = line_number
	glb_source_file_name = source_file_name

	print(start_str, src_str)

def do_process_event(param_dict):
	sample	   = param_dict["sample"]
	raw_buf	   = param_dict["raw_buf"]
	comm	   = param_dict["comm"]
	name	   = param_dict["ev_name"]
	# Unused fields:
	# callchain  = param_dict["callchain"]
	# brstack    = param_dict["brstack"]
	# brstacksym = param_dict["brstacksym"]
	# event_attr = param_dict["attr"]

	# Symbol and dso info are not always resolved
	dso    = get_optional(param_dict, "dso")
	symbol = get_optional(param_dict, "symbol")

	cpu = sample["cpu"]
	if cpu in glb_switch_str:
		print(glb_switch_str[cpu])
		del glb_switch_str[cpu]

	if name.startswith("instructions"):
		if glb_src:
			print_srccode(comm, param_dict, sample, symbol, dso, True)
		else:
			print_instructions_start(comm, sample)
			print_common_ip(param_dict, sample, symbol, dso)
	elif name.startswith("branches"):
		if glb_src:
			print_srccode(comm, param_dict, sample, symbol, dso, False)
		else:
			print_common_start(comm, sample, name)
			print_common_ip(param_dict, sample, symbol, dso)
	elif name == "ptwrite":
		print_common_start(comm, sample, name)
		print_ptwrite(raw_buf)
		print_common_ip(param_dict, sample, symbol, dso)
	elif name == "cbr":
		print_common_start(comm, sample, name)
		print_cbr(raw_buf)
		print_common_ip(param_dict, sample, symbol, dso)
	elif name == "mwait":
		print_common_start(comm, sample, name)
		print_mwait(raw_buf)
		print_common_ip(param_dict, sample, symbol, dso)
	elif name == "pwre":
		print_common_start(comm, sample, name)
		print_pwre(raw_buf)
		print_common_ip(param_dict, sample, symbol, dso)
	elif name == "exstop":
		print_common_start(comm, sample, name)
		print_exstop(raw_buf)
		print_common_ip(param_dict, sample, symbol, dso)
	elif name == "pwrx":
		print_common_start(comm, sample, name)
		print_pwrx(raw_buf)
		print_common_ip(param_dict, sample, symbol, dso)
	elif name == "psb":
		print_common_start(comm, sample, name)
		print_psb(raw_buf)
		print_common_ip(param_dict, sample, symbol, dso)
	elif name == "evt":
		print_common_start(comm, sample, name)
		print_evt(raw_buf)
		print_common_ip(param_dict, sample, symbol, dso)
	elif name == "iflag":
		print_common_start(comm, sample, name)
		print_iflag(raw_buf)
		print_common_ip(param_dict, sample, symbol, dso)
	else:
		print_common_start(comm, sample, name)
		print_common_ip(param_dict, sample, symbol, dso)

def interleave_events(param_dict):
	global glb_cpu
	global glb_time
	global glb_output
	global glb_output_pos

	sample  = param_dict["sample"]
	glb_cpu = sample["cpu"]
	ts      = sample["time"]

	if glb_time != ts:
		glb_time = ts
		flush_stashed_output()

	glb_output_pos = 0
	with contextlib.redirect_stdout(io.StringIO()) as glb_output:
		do_process_event(param_dict)

	stash_output()

def process_event(param_dict):
	try:
		if glb_args.interleave:
			interleave_events(param_dict)
		else:
			do_process_event(param_dict)
	except broken_pipe_exception:
		# Stop python printing broken pipe errors and traceback
		sys.stdout = open(os.devnull, 'w')
		sys.exit(1)

def auxtrace_error(typ, code, cpu, pid, tid, ip, ts, msg, cpumode, *x):
	if glb_args.interleave:
		flush_stashed_output()
	if len(x) >= 2 and x[0]:
		machine_pid = x[0]
		vcpu = x[1]
	else:
		machine_pid = 0
		vcpu = -1
	try:
		if machine_pid:
			print("VM:%5d VCPU:%03d %16s %5u/%-5u [%03u] %9u.%09u  error type %u code %u: %s ip 0x%16x" %
				(machine_pid, vcpu, "Trace error", pid, tid, cpu, ts / 1000000000, ts %1000000000, typ, code, msg, ip))
		else:
			print("%16s %5u/%-5u [%03u] %9u.%09u  error type %u code %u: %s ip 0x%16x" %
				("Trace error", pid, tid, cpu, ts / 1000000000, ts %1000000000, typ, code, msg, ip))
	except broken_pipe_exception:
		# Stop python printing broken pipe errors and traceback
		sys.stdout = open(os.devnull, 'w')
		sys.exit(1)

def context_switch(ts, cpu, pid, tid, np_pid, np_tid, machine_pid, out, out_preempt, *x):
	if glb_args.interleave:
		flush_stashed_output()
	if out:
		out_str = "Switch out "
	else:
		out_str = "Switch In  "
	if out_preempt:
		preempt_str = "preempt"
	else:
		preempt_str = ""
	if len(x) >= 2 and x[0]:
		machine_pid = x[0]
		vcpu = x[1]
	else:
		vcpu = None;
	if machine_pid == -1:
		machine_str = ""
	elif vcpu is None:
		machine_str = "machine PID %d" % machine_pid
	else:
		machine_str = "machine PID %d VCPU %d" % (machine_pid, vcpu)
	switch_str = "%16s %5d/%-5d [%03u] %9u.%09u %5d/%-5d %s %s" % \
		(out_str, pid, tid, cpu, ts / 1000000000, ts %1000000000, np_pid, np_tid, machine_str, preempt_str)
	if glb_args.all_switch_events:
		print(switch_str)
	else:
		global glb_switch_str
		glb_switch_str[cpu] = switch_str



# EventClass.py
# SPDX-License-Identifier: GPL-2.0
#
# This is a library defining some events types classes, which could
# be used by other scripts to analyzing the perf samples.
#
# Currently there are just a few classes defined for examples,
# PerfEvent is the base class for all perf event sample, PebsEvent
# is a HW base Intel x86 PEBS event, and user could add more SW/HW
# event classes based on requirements.
from __future__ import print_function

import struct

# Event types, user could add more here
EVTYPE_GENERIC  = 0
EVTYPE_PEBS     = 1     # Basic PEBS event
EVTYPE_PEBS_LL  = 2     # PEBS event with load latency info
EVTYPE_IBS      = 3

#
# Currently we don't have good way to tell the event type, but by
# the size of raw buffer, raw PEBS event with load latency data's
# size is 176 bytes, while the pure PEBS event's size is 144 bytes.
#
def create_event(name, comm, dso, symbol, raw_buf):
        if (len(raw_buf) == 144):
                event = PebsEvent(name, comm, dso, symbol, raw_buf)
        elif (len(raw_buf) == 176):
                event = PebsNHM(name, comm, dso, symbol, raw_buf)
        else:
                event = PerfEvent(name, comm, dso, symbol, raw_buf)

        return event

class PerfEvent(object):
        event_num = 0
        def __init__(self, name, comm, dso, symbol, raw_buf, ev_type=EVTYPE_GENERIC):
                self.name       = name
                self.comm       = comm
                self.dso        = dso
                self.symbol     = symbol
                self.raw_buf    = raw_buf
                self.ev_type    = ev_type
                PerfEvent.event_num += 1

        def show(self):
                print("PMU event: name=%12s, symbol=%24s, comm=%8s, dso=%12s" %
                      (self.name, self.symbol, self.comm, self.dso))

#
# Basic Intel PEBS (Precise Event-based Sampling) event, whose raw buffer
# contains the context info when that event happened: the EFLAGS and
# linear IP info, as well as all the registers.
#
class PebsEvent(PerfEvent):
        pebs_num = 0
        def __init__(self, name, comm, dso, symbol, raw_buf, ev_type=EVTYPE_PEBS):
                tmp_buf=raw_buf[0:80]
                flags, ip, ax, bx, cx, dx, si, di, bp, sp = struct.unpack('QQQQQQQQQQ', tmp_buf)
                self.flags = flags
                self.ip    = ip
                self.ax    = ax
                self.bx    = bx
                self.cx    = cx
                self.dx    = dx
                self.si    = si
                self.di    = di
                self.bp    = bp
                self.sp    = sp

                PerfEvent.__init__(self, name, comm, dso, symbol, raw_buf, ev_type)
                PebsEvent.pebs_num += 1
                del tmp_buf

#
# Intel Nehalem and Westmere support PEBS plus Load Latency info which lie
# in the four 64 bit words write after the PEBS data:
#       Status: records the IA32_PERF_GLOBAL_STATUS register value
#       DLA:    Data Linear Address (EIP)
#       DSE:    Data Source Encoding, where the latency happens, hit or miss
#               in L1/L2/L3 or IO operations
#       LAT:    the actual latency in cycles
#
class PebsNHM(PebsEvent):
        pebs_nhm_num = 0
        def __init__(self, name, comm, dso, symbol, raw_buf, ev_type=EVTYPE_PEBS_LL):
                tmp_buf=raw_buf[144:176]
                status, dla, dse, lat = struct.unpack('QQQQ', tmp_buf)
                self.status = status
                self.dla = dla
                self.dse = dse
                self.lat = lat

                PebsEvent.__init__(self, name, comm, dso, symbol, raw_buf, ev_type)
                PebsNHM.pebs_nhm_num += 1
                del tmp_buf



# system call counts, by pid
# (c) 2010, Tom Zanussi <tzanussi@gmail.com>
# Licensed under the terms of the GNU GPL License version 2
#
# Displays system-wide system call totals, broken down by syscall.
# If a [comm] arg is specified, only syscalls called by [comm] are displayed.

from __future__ import print_function

import os, sys

sys.path.append(os.environ['PERF_EXEC_PATH'] + \
	'/scripts/python/Perf-Trace-Util/lib/Perf/Trace')

from perf_trace_context import *
from Core import *
from Util import syscall_name

usage = "perf script -s syscall-counts-by-pid.py [comm]\n";

for_comm = None
for_pid = None

if len(sys.argv) > 2:
	sys.exit(usage)

if len(sys.argv) > 1:
	try:
		for_pid = int(sys.argv[1])
	except:
		for_comm = sys.argv[1]

syscalls = autodict()

def trace_begin():
	print("Press control+C to stop and show the summary")

def trace_end():
	print_syscall_totals()

def raw_syscalls__sys_enter(event_name, context, common_cpu,
		common_secs, common_nsecs, common_pid, common_comm,
		common_callchain, id, args):
	if (for_comm and common_comm != for_comm) or \
		(for_pid and common_pid != for_pid ):
		return
	try:
		syscalls[common_comm][common_pid][id] += 1
	except TypeError:
		syscalls[common_comm][common_pid][id] = 1

def syscalls__sys_enter(event_name, context, common_cpu,
		common_secs, common_nsecs, common_pid, common_comm,
		id, args):
	raw_syscalls__sys_enter(**locals())

def print_syscall_totals():
	if for_comm is not None:
		print("\nsyscall events for %s:\n" % (for_comm))
	else:
		print("\nsyscall events by comm/pid:\n")

	print("%-40s  %10s" % ("comm [pid]/syscalls", "count"))
	print("%-40s  %10s" % ("----------------------------------------",
				"----------"))

	comm_keys = syscalls.keys()
	for comm in comm_keys:
		pid_keys = syscalls[comm].keys()
		for pid in pid_keys:
			print("\n%s [%d]" % (comm, pid))
			id_keys = syscalls[comm][pid].keys()
			for id, val in sorted(syscalls[comm][pid].items(),
				key = lambda kv: (kv[1], kv[0]), reverse = True):
				print("  %-38s  %10d" % (syscall_name(id), val))



#!/usr/bin/env python3
# SPDX-License-Identifier: GPL-2.0
#
# Run a perf script command multiple times in parallel, using perf script
# options --cpu and --time so that each job processes a different chunk
# of the data.
#
# Copyright (c) 2024, Intel Corporation.

import subprocess
import argparse
import pathlib
import shlex
import time
import copy
import sys
import os
import re

glb_prog_name = "parallel-perf.py"
glb_min_interval = 10.0
glb_min_samples = 64

class Verbosity():

	def __init__(self, quiet=False, verbose=False, debug=False):
		self.normal    = True
		self.verbose   = verbose
		self.debug     = debug
		self.self_test = True
		if self.debug:
			self.verbose = True
		if self.verbose:
			quiet = False
		if quiet:
			self.normal = False

# Manage work (Start/Wait/Kill), as represented by a subprocess.Popen command
class Work():

	def __init__(self, cmd, pipe_to, output_dir="."):
		self.popen = None
		self.consumer = None
		self.cmd = cmd
		self.pipe_to = pipe_to
		self.output_dir = output_dir
		self.cmdout_name = f"{output_dir}/cmd.txt"
		self.stdout_name = f"{output_dir}/out.txt"
		self.stderr_name = f"{output_dir}/err.txt"

	def Command(self):
		sh_cmd = [ shlex.quote(x) for x in self.cmd ]
		return " ".join(self.cmd)

	def Stdout(self):
		return open(self.stdout_name, "w")

	def Stderr(self):
		return open(self.stderr_name, "w")

	def CreateOutputDir(self):
		pathlib.Path(self.output_dir).mkdir(parents=True, exist_ok=True)

	def Start(self):
		if self.popen:
			return
		self.CreateOutputDir()
		with open(self.cmdout_name, "w") as f:
			f.write(self.Command())
			f.write("\n")
		stdout = self.Stdout()
		stderr = self.Stderr()
		if self.pipe_to:
			self.popen = subprocess.Popen(self.cmd, stdout=subprocess.PIPE, stderr=stderr)
			args = shlex.split(self.pipe_to)
			self.consumer = subprocess.Popen(args, stdin=self.popen.stdout, stdout=stdout, stderr=stderr)
		else:
			self.popen = subprocess.Popen(self.cmd, stdout=stdout, stderr=stderr)

	def RemoveEmptyErrFile(self):
		if os.path.exists(self.stderr_name):
			if os.path.getsize(self.stderr_name) == 0:
				os.unlink(self.stderr_name)

	def Errors(self):
		if os.path.exists(self.stderr_name):
			if os.path.getsize(self.stderr_name) != 0:
				return [ f"Non-empty error file {self.stderr_name}" ]
		return []

	def TidyUp(self):
		self.RemoveEmptyErrFile()

	def RawPollWait(self, p, wait):
		if wait:
			return p.wait()
		return p.poll()

	def Poll(self, wait=False):
		if not self.popen:
			return None
		result = self.RawPollWait(self.popen, wait)
		if self.consumer:
			res = result
			result = self.RawPollWait(self.consumer, wait)
			if result != None and res == None:
				self.popen.kill()
				result = None
			elif result == 0 and res != None and res != 0:
				result = res
		if result != None:
			self.TidyUp()
		return result

	def Wait(self):
		return self.Poll(wait=True)

	def Kill(self):
		if not self.popen:
			return
		self.popen.kill()
		if self.consumer:
			self.consumer.kill()

def KillWork(worklist, verbosity):
	for w in worklist:
		w.Kill()
	for w in worklist:
		w.Wait()

def NumberOfCPUs():
	return os.sysconf("SC_NPROCESSORS_ONLN")

def NanoSecsToSecsStr(x):
	if x == None:
		return ""
	x = str(x)
	if len(x) < 10:
		x = "0" * (10 - len(x)) + x
	return x[:len(x) - 9] + "." + x[-9:]

def InsertOptionAfter(cmd, option, after):
	try:
		pos = cmd.index(after)
		cmd.insert(pos + 1, option)
	except:
		cmd.append(option)

def CreateWorkList(cmd, pipe_to, output_dir, cpus, time_ranges_by_cpu):
	max_len = len(str(cpus[-1]))
	cpu_dir_fmt = f"cpu-%.{max_len}u"
	worklist = []
	pos = 0
	for cpu in cpus:
		if cpu >= 0:
			cpu_dir = os.path.join(output_dir, cpu_dir_fmt % cpu)
			cpu_option = f"--cpu={cpu}"
		else:
			cpu_dir = output_dir
			cpu_option = None

		tr_dir_fmt = "time-range"

		if len(time_ranges_by_cpu) > 1:
			time_ranges = time_ranges_by_cpu[pos]
			tr_dir_fmt += f"-{pos}"
			pos += 1
		else:
			time_ranges = time_ranges_by_cpu[0]

		max_len = len(str(len(time_ranges)))
		tr_dir_fmt += f"-%.{max_len}u"

		i = 0
		for r in time_ranges:
			if r == [None, None]:
				time_option = None
				work_output_dir = cpu_dir
			else:
				time_option = "--time=" + NanoSecsToSecsStr(r[0]) + "," + NanoSecsToSecsStr(r[1])
				work_output_dir = os.path.join(cpu_dir, tr_dir_fmt % i)
				i += 1
			work_cmd = list(cmd)
			if time_option != None:
				InsertOptionAfter(work_cmd, time_option, "script")
			if cpu_option != None:
				InsertOptionAfter(work_cmd, cpu_option, "script")
			w = Work(work_cmd, pipe_to, work_output_dir)
			worklist.append(w)
	return worklist

def DoRunWork(worklist, nr_jobs, verbosity):
	nr_to_do = len(worklist)
	not_started = list(worklist)
	running = []
	done = []
	chg = False
	while True:
		nr_done = len(done)
		if chg and verbosity.normal:
			nr_run = len(running)
			print(f"\rThere are {nr_to_do} jobs: {nr_done} completed, {nr_run} running", flush=True, end=" ")
			if verbosity.verbose:
				print()
			chg = False
		if nr_done == nr_to_do:
			break
		while len(running) < nr_jobs and len(not_started):
			w = not_started.pop(0)
			running.append(w)
			if verbosity.verbose:
				print("Starting:", w.Command())
			w.Start()
			chg = True
		if len(running):
			time.sleep(0.1)
		finished = []
		not_finished = []
		while len(running):
			w = running.pop(0)
			r = w.Poll()
			if r == None:
				not_finished.append(w)
				continue
			if r == 0:
				if verbosity.verbose:
					print("Finished:", w.Command())
				finished.append(w)
				chg = True
				continue
			if verbosity.normal and not verbosity.verbose:
				print()
			print("Job failed!\n    return code:", r, "\n    command:    ", w.Command())
			if w.pipe_to:
				print("    piped to:   ", w.pipe_to)
			print("Killing outstanding jobs")
			KillWork(not_finished, verbosity)
			KillWork(running, verbosity)
			return False
		running = not_finished
		done += finished
	errorlist = []
	for w in worklist:
		errorlist += w.Errors()
	if len(errorlist):
		print("Errors:")
		for e in errorlist:
			print(e)
	elif verbosity.normal:
		print("\r"," "*50, "\rAll jobs finished successfully", flush=True)
	return True

def RunWork(worklist, nr_jobs=NumberOfCPUs(), verbosity=Verbosity()):
	try:
		return DoRunWork(worklist, nr_jobs, verbosity)
	except:
		for w in worklist:
			w.Kill()
		raise
	return True

def ReadHeader(perf, file_name):
	return subprocess.Popen([perf, "script", "--header-only", "--input", file_name], stdout=subprocess.PIPE).stdout.read().decode("utf-8")

def ParseHeader(hdr):
	result = {}
	lines = hdr.split("\n")
	for line in lines:
		if ":" in line and line[0] == "#":
			pos = line.index(":")
			name = line[1:pos-1].strip()
			value = line[pos+1:].strip()
			if name in result:
				orig_name = name
				nr = 2
				while True:
					name = f"{orig_name} {nr}"
					if name not in result:
						break
					nr += 1
			result[name] = value
	return result

def HeaderField(hdr_dict, hdr_fld):
	if hdr_fld not in hdr_dict:
		raise Exception(f"'{hdr_fld}' missing from header information")
	return hdr_dict[hdr_fld]

# Represent the position of an option within a command string
# and provide the option value and/or remove the option
class OptPos():

	def Init(self, opt_element=-1, value_element=-1, opt_pos=-1, value_pos=-1, error=None):
		self.opt_element = opt_element		# list element that contains option
		self.value_element = value_element	# list element that contains option value
		self.opt_pos = opt_pos			# string position of option
		self.value_pos = value_pos		# string position of value
		self.error = error			# error message string

	def __init__(self, args, short_name, long_name, default=None):
		self.args = list(args)
		self.default = default
		n = 2 + len(long_name)
		m = len(short_name)
		pos = -1
		for opt in args:
			pos += 1
			if m and opt[:2] == f"-{short_name}":
				if len(opt) == 2:
					if pos + 1 < len(args):
						self.Init(pos, pos + 1, 0, 0)
					else:
						self.Init(error = f"-{short_name} option missing value")
				else:
					self.Init(pos, pos, 0, 2)
				return
			if opt[:n] == f"--{long_name}":
				if len(opt) == n:
					if pos + 1 < len(args):
						self.Init(pos, pos + 1, 0, 0)
					else:
						self.Init(error = f"--{long_name} option missing value")
				elif opt[n] == "=":
					self.Init(pos, pos, 0, n + 1)
				else:
					self.Init(error = f"--{long_name} option expected '='")
				return
			if m and opt[:1] == "-" and opt[:2] != "--" and short_name in opt:
				ipos = opt.index(short_name)
				if "-" in opt[1:]:
					hpos = opt[1:].index("-")
					if hpos < ipos:
						continue
				if ipos + 1 == len(opt):
					if pos + 1 < len(args):
						self.Init(pos, pos + 1, ipos, 0)
					else:
						self.Init(error = f"-{short_name} option missing value")
				else:
					self.Init(pos, pos, ipos, ipos + 1)
				return
		self.Init()

	def Value(self):
		if self.opt_element >= 0:
			if self.opt_element != self.value_element:
				return self.args[self.value_element]
			else:
				return self.args[self.value_element][self.value_pos:]
		return self.default

	def Remove(self, args):
		if self.opt_element == -1:
			return
		if self.opt_element != self.value_element:
			del args[self.value_element]
		if self.opt_pos:
			args[self.opt_element] = args[self.opt_element][:self.opt_pos]
		else:
			del args[self.opt_element]

def DetermineInputFileName(cmd):
	p = OptPos(cmd, "i", "input", "perf.data")
	if p.error:
		raise Exception(f"perf command {p.error}")
	file_name = p.Value()
	if not os.path.exists(file_name):
		raise Exception(f"perf command input file '{file_name}' not found")
	return file_name

def ReadOption(args, short_name, long_name, err_prefix, remove=False):
	p = OptPos(args, short_name, long_name)
	if p.error:
		raise Exception(f"{err_prefix}{p.error}")
	value = p.Value()
	if remove:
		p.Remove(args)
	return value

def ExtractOption(args, short_name, long_name, err_prefix):
	return ReadOption(args, short_name, long_name, err_prefix, True)

def ReadPerfOption(args, short_name, long_name):
	return ReadOption(args, short_name, long_name, "perf command ")

def ExtractPerfOption(args, short_name, long_name):
	return ExtractOption(args, short_name, long_name, "perf command ")

def PerfDoubleQuickCommands(cmd, file_name):
	cpu_str = ReadPerfOption(cmd, "C", "cpu")
	time_str = ReadPerfOption(cmd, "", "time")
	# Use double-quick sampling to determine trace data density
	times_cmd = ["perf", "script", "--ns", "--input", file_name, "--itrace=qqi"]
	if cpu_str != None and cpu_str != "":
		times_cmd.append(f"--cpu={cpu_str}")
	if time_str != None and time_str != "":
		times_cmd.append(f"--time={time_str}")
	cnts_cmd = list(times_cmd)
	cnts_cmd.append("-Fcpu")
	times_cmd.append("-Fcpu,time")
	return cnts_cmd, times_cmd

class CPUTimeRange():
	def __init__(self, cpu):
		self.cpu = cpu
		self.sample_cnt = 0
		self.time_ranges = None
		self.interval = 0
		self.interval_remaining = 0
		self.remaining = 0
		self.tr_pos = 0

def CalcTimeRangesByCPU(line, cpu, cpu_time_ranges, max_time):
	cpu_time_range = cpu_time_ranges[cpu]
	cpu_time_range.remaining -= 1
	cpu_time_range.interval_remaining -= 1
	if cpu_time_range.remaining == 0:
		cpu_time_range.time_ranges[cpu_time_range.tr_pos][1] = max_time
		return
	if cpu_time_range.interval_remaining == 0:
		time = TimeVal(line[1][:-1], 0)
		time_ranges = cpu_time_range.time_ranges
		time_ranges[cpu_time_range.tr_pos][1] = time - 1
		time_ranges.append([time, max_time])
		cpu_time_range.tr_pos += 1
		cpu_time_range.interval_remaining = cpu_time_range.interval

def CountSamplesByCPU(line, cpu, cpu_time_ranges):
	try:
		cpu_time_ranges[cpu].sample_cnt += 1
	except:
		print("exception")
		print("cpu", cpu)
		print("len(cpu_time_ranges)", len(cpu_time_ranges))
		raise

def ProcessCommandOutputLines(cmd, per_cpu, fn, *x):
	# Assume CPU number is at beginning of line and enclosed by []
	pat = re.compile(r"\s*\[[0-9]+\]")
	p = subprocess.Popen(cmd, stdout=subprocess.PIPE)
	while True:
		line = p.stdout.readline()
		if line:
			line = line.decode("utf-8")
			if pat.match(line):
				line = line.split()
				if per_cpu:
					# Assumes CPU number is enclosed by []
					cpu = int(line[0][1:-1])
				else:
					cpu = 0
				fn(line, cpu, *x)
		else:
			break
	p.wait()

def IntersectTimeRanges(new_time_ranges, time_ranges):
	pos = 0
	new_pos = 0
	# Can assume len(time_ranges) != 0 and len(new_time_ranges) != 0
	# Note also, there *must* be at least one intersection.
	while pos < len(time_ranges) and new_pos < len(new_time_ranges):
		# new end < old start => no intersection, remove new
		if new_time_ranges[new_pos][1] < time_ranges[pos][0]:
			del new_time_ranges[new_pos]
			continue
		# new start > old end => no intersection, check next
		if new_time_ranges[new_pos][0] > time_ranges[pos][1]:
			pos += 1
			if pos < len(time_ranges):
				continue
			# no next, so remove remaining
			while new_pos < len(new_time_ranges):
				del new_time_ranges[new_pos]
			return
		# Found an intersection
		# new start < old start => adjust new start = old start
		if new_time_ranges[new_pos][0] < time_ranges[pos][0]:
			new_time_ranges[new_pos][0] = time_ranges[pos][0]
		# new end > old end => keep the overlap, insert the remainder
		if new_time_ranges[new_pos][1] > time_ranges[pos][1]:
			r = [ time_ranges[pos][1] + 1, new_time_ranges[new_pos][1] ]
			new_time_ranges[new_pos][1] = time_ranges[pos][1]
			new_pos += 1
			new_time_ranges.insert(new_pos, r)
			continue
		# new [start, end] is within old [start, end]
		new_pos += 1

def SplitTimeRangesByTraceDataDensity(time_ranges, cpus, nr, cmd, file_name, per_cpu, min_size, min_interval, verbosity):
	if verbosity.normal:
		print("\rAnalyzing...", flush=True, end=" ")
		if verbosity.verbose:
			print()
	cnts_cmd, times_cmd = PerfDoubleQuickCommands(cmd, file_name)

	nr_cpus = cpus[-1] + 1 if per_cpu else 1
	if per_cpu:
		nr_cpus = cpus[-1] + 1
		cpu_time_ranges = [ CPUTimeRange(cpu) for cpu in range(nr_cpus) ]
	else:
		nr_cpus = 1
		cpu_time_ranges = [ CPUTimeRange(-1) ]

	if verbosity.debug:
		print("nr_cpus", nr_cpus)
		print("cnts_cmd", cnts_cmd)
		print("times_cmd", times_cmd)

	# Count the number of "double quick" samples per CPU
	ProcessCommandOutputLines(cnts_cmd, per_cpu, CountSamplesByCPU, cpu_time_ranges)

	tot = 0
	mx = 0
	for cpu_time_range in cpu_time_ranges:
		cnt = cpu_time_range.sample_cnt
		tot += cnt
		if cnt > mx:
			mx = cnt
		if verbosity.debug:
			print("cpu:", cpu_time_range.cpu, "sample_cnt", cnt)

	if min_size < 1:
		min_size = 1

	if mx < min_size:
		# Too little data to be worth splitting
		if verbosity.debug:
			print("Too little data to split by time")
		if nr == 0:
			nr = 1
		return [ SplitTimeRangesIntoN(time_ranges, nr, min_interval) ]

	if nr:
		divisor = nr
		min_size = 1
	else:
		divisor = NumberOfCPUs()

	interval = int(round(tot / divisor, 0))
	if interval < min_size:
		interval = min_size

	if verbosity.debug:
		print("divisor", divisor)
		print("min_size", min_size)
		print("interval", interval)

	min_time = time_ranges[0][0]
	max_time = time_ranges[-1][1]

	for cpu_time_range in cpu_time_ranges:
		cnt = cpu_time_range.sample_cnt
		if cnt == 0:
			cpu_time_range.time_ranges = copy.deepcopy(time_ranges)
			continue
		# Adjust target interval for CPU to give approximately equal interval sizes
		# Determine number of intervals, rounding to nearest integer
		n = int(round(cnt / interval, 0))
		if n < 1:
			n = 1
		# Determine interval size, rounding up
		d, m = divmod(cnt, n)
		if m:
			d += 1
		cpu_time_range.interval = d
		cpu_time_range.interval_remaining = d
		cpu_time_range.remaining = cnt
		# Init. time ranges for each CPU with the start time
		cpu_time_range.time_ranges = [ [min_time, max_time] ]

	# Set time ranges so that the same number of "double quick" samples
	# will fall into each time range.
	ProcessCommandOutputLines(times_cmd, per_cpu, CalcTimeRangesByCPU, cpu_time_ranges, max_time)

	for cpu_time_range in cpu_time_ranges:
		if cpu_time_range.sample_cnt:
			IntersectTimeRanges(cpu_time_range.time_ranges, time_ranges)

	return [cpu_time_ranges[cpu].time_ranges for cpu in cpus]

def SplitSingleTimeRangeIntoN(time_range, n):
	if n <= 1:
		return [time_range]
	start = time_range[0]
	end   = time_range[1]
	duration = int((end - start + 1) / n)
	if duration < 1:
		return [time_range]
	time_ranges = []
	for i in range(n):
		time_ranges.append([start, start + duration - 1])
		start += duration
	time_ranges[-1][1] = end
	return time_ranges

def TimeRangeDuration(r):
	return r[1] - r[0] + 1

def TotalDuration(time_ranges):
	duration = 0
	for r in time_ranges:
		duration += TimeRangeDuration(r)
	return duration

def SplitTimeRangesByInterval(time_ranges, interval):
	new_ranges = []
	for r in time_ranges:
		duration = TimeRangeDuration(r)
		n = duration / interval
		n = int(round(n, 0))
		new_ranges += SplitSingleTimeRangeIntoN(r, n)
	return new_ranges

def SplitTimeRangesIntoN(time_ranges, n, min_interval):
	if n <= len(time_ranges):
		return time_ranges
	duration = TotalDuration(time_ranges)
	interval = duration / n
	if interval < min_interval:
		interval = min_interval
	return SplitTimeRangesByInterval(time_ranges, interval)

def RecombineTimeRanges(tr):
	new_tr = copy.deepcopy(tr)
	n = len(new_tr)
	i = 1
	while i < len(new_tr):
		# if prev end + 1 == cur start, combine them
		if new_tr[i - 1][1] + 1 == new_tr[i][0]:
			new_tr[i][0] = new_tr[i - 1][0]
			del new_tr[i - 1]
		else:
			i += 1
	return new_tr

def OpenTimeRangeEnds(time_ranges, min_time, max_time):
	if time_ranges[0][0] <= min_time:
		time_ranges[0][0] = None
	if time_ranges[-1][1] >= max_time:
		time_ranges[-1][1] = None

def BadTimeStr(time_str):
	raise Exception(f"perf command bad time option: '{time_str}'\nCheck also 'time of first sample' and 'time of last sample' in perf script --header-only")

def ValidateTimeRanges(time_ranges, time_str):
	n = len(time_ranges)
	for i in range(n):
		start = time_ranges[i][0]
		end   = time_ranges[i][1]
		if i != 0 and start <= time_ranges[i - 1][1]:
			BadTimeStr(time_str)
		if start > end:
			BadTimeStr(time_str)

def TimeVal(s, dflt):
	s = s.strip()
	if s == "":
		return dflt
	a = s.split(".")
	if len(a) > 2:
		raise Exception(f"Bad time value'{s}'")
	x = int(a[0])
	if x < 0:
		raise Exception("Negative time not allowed")
	x *= 1000000000
	if len(a) > 1:
		x += int((a[1] + "000000000")[:9])
	return x

def BadCPUStr(cpu_str):
	raise Exception(f"perf command bad cpu option: '{cpu_str}'\nCheck also 'nrcpus avail' in perf script --header-only")

def ParseTimeStr(time_str, min_time, max_time):
	if time_str == None or time_str == "":
		return [[min_time, max_time]]
	time_ranges = []
	for r in time_str.split():
		a = r.split(",")
		if len(a) != 2:
			BadTimeStr(time_str)
		try:
			start = TimeVal(a[0], min_time)
			end   = TimeVal(a[1], max_time)
		except:
			BadTimeStr(time_str)
		time_ranges.append([start, end])
	ValidateTimeRanges(time_ranges, time_str)
	return time_ranges

def ParseCPUStr(cpu_str, nr_cpus):
	if cpu_str == None or cpu_str == "":
		return [-1]
	cpus = []
	for r in cpu_str.split(","):
		a = r.split("-")
		if len(a) < 1 or len(a) > 2:
			BadCPUStr(cpu_str)
		try:
			start = int(a[0].strip())
			if len(a) > 1:
				end = int(a[1].strip())
			else:
				end = start
		except:
			BadCPUStr(cpu_str)
		if start < 0 or end < 0 or end < start or end >= nr_cpus:
			BadCPUStr(cpu_str)
		cpus.extend(range(start, end + 1))
	cpus = list(set(cpus)) # Remove duplicates
	cpus.sort()
	return cpus

class ParallelPerf():

	def __init__(self, a):
		for arg_name in vars(a):
			setattr(self, arg_name, getattr(a, arg_name))
		self.orig_nr = self.nr
		self.orig_cmd = list(self.cmd)
		self.perf = self.cmd[0]
		if os.path.exists(self.output_dir):
			raise Exception(f"Output '{self.output_dir}' already exists")
		if self.jobs < 0 or self.nr < 0 or self.interval < 0:
			raise Exception("Bad options (negative values): try -h option for help")
		if self.nr != 0 and self.interval != 0:
			raise Exception("Cannot specify number of time subdivisions and time interval")
		if self.jobs == 0:
			self.jobs = NumberOfCPUs()
		if self.nr == 0 and self.interval == 0:
			if self.per_cpu:
				self.nr = 1
			else:
				self.nr = self.jobs

	def Init(self):
		if self.verbosity.debug:
			print("cmd", self.cmd)
		self.file_name = DetermineInputFileName(self.cmd)
		self.hdr = ReadHeader(self.perf, self.file_name)
		self.hdr_dict = ParseHeader(self.hdr)
		self.cmd_line = HeaderField(self.hdr_dict, "cmdline")

	def ExtractTimeInfo(self):
		self.min_time = TimeVal(HeaderField(self.hdr_dict, "time of first sample"), 0)
		self.max_time = TimeVal(HeaderField(self.hdr_dict, "time of last sample"), 0)
		self.time_str = ExtractPerfOption(self.cmd, "", "time")
		self.time_ranges = ParseTimeStr(self.time_str, self.min_time, self.max_time)
		if self.verbosity.debug:
			print("time_ranges", self.time_ranges)

	def ExtractCPUInfo(self):
		if self.per_cpu:
			nr_cpus = int(HeaderField(self.hdr_dict, "nrcpus avail"))
			self.cpu_str = ExtractPerfOption(self.cmd, "C", "cpu")
			if self.cpu_str == None or self.cpu_str == "":
				self.cpus = [ x for x in range(nr_cpus) ]
			else:
				self.cpus = ParseCPUStr(self.cpu_str, nr_cpus)
		else:
			self.cpu_str = None
			self.cpus = [-1]
		if self.verbosity.debug:
			print("cpus", self.cpus)

	def IsIntelPT(self):
		return self.cmd_line.find("intel_pt") >= 0

	def SplitTimeRanges(self):
		if self.IsIntelPT() and self.interval == 0:
			self.split_time_ranges_for_each_cpu = \
				SplitTimeRangesByTraceDataDensity(self.time_ranges, self.cpus, self.orig_nr,
								  self.orig_cmd, self.file_name, self.per_cpu,
								  self.min_size, self.min_interval, self.verbosity)
		elif self.nr:
			self.split_time_ranges_for_each_cpu = [ SplitTimeRangesIntoN(self.time_ranges, self.nr, self.min_interval) ]
		else:
			self.split_time_ranges_for_each_cpu = [ SplitTimeRangesByInterval(self.time_ranges, self.interval) ]

	def CheckTimeRanges(self):
		for tr in self.split_time_ranges_for_each_cpu:
			# Re-combined time ranges should be the same
			new_tr = RecombineTimeRanges(tr)
			if new_tr != self.time_ranges:
				if self.verbosity.debug:
					print("tr", tr)
					print("new_tr", new_tr)
				raise Exception("Self test failed!")

	def OpenTimeRangeEnds(self):
		for time_ranges in self.split_time_ranges_for_each_cpu:
			OpenTimeRangeEnds(time_ranges, self.min_time, self.max_time)

	def CreateWorkList(self):
		self.worklist = CreateWorkList(self.cmd, self.pipe_to, self.output_dir, self.cpus, self.split_time_ranges_for_each_cpu)

	def PerfDataRecordedPerCPU(self):
		if "--per-thread" in self.cmd_line.split():
			return False
		return True

	def DefaultToPerCPU(self):
		# --no-per-cpu option takes precedence
		if self.no_per_cpu:
			return False
		if not self.PerfDataRecordedPerCPU():
			return False
		# Default to per-cpu for Intel PT data that was recorded per-cpu,
		# because decoding can be done for each CPU separately.
		if self.IsIntelPT():
			return True
		return False

	def Config(self):
		self.Init()
		self.ExtractTimeInfo()
		if not self.per_cpu:
			self.per_cpu = self.DefaultToPerCPU()
		if self.verbosity.debug:
			print("per_cpu", self.per_cpu)
		self.ExtractCPUInfo()
		self.SplitTimeRanges()
		if self.verbosity.self_test:
			self.CheckTimeRanges()
		# Prefer open-ended time range to starting / ending with min_time / max_time resp.
		self.OpenTimeRangeEnds()
		self.CreateWorkList()

	def Run(self):
		if self.dry_run:
			print(len(self.worklist),"jobs:")
			for w in self.worklist:
				print(w.Command())
			return True
		result = RunWork(self.worklist, self.jobs, verbosity=self.verbosity)
		if self.verbosity.verbose:
			print(glb_prog_name, "done")
		return result

def RunParallelPerf(a):
	pp = ParallelPerf(a)
	pp.Config()
	return pp.Run()

def Main(args):
	ap = argparse.ArgumentParser(
		prog=glb_prog_name, formatter_class = argparse.RawDescriptionHelpFormatter,
		description =
"""
Run a perf script command multiple times in parallel, using perf script options
--cpu and --time so that each job processes a different chunk of the data.
""",
		epilog =
"""
Follow the options by '--' and then the perf script command e.g.

	$ perf record -a -- sleep 10
	$ parallel-perf.py --nr=4 -- perf script --ns
	All jobs finished successfully
	$ tree parallel-perf-output/
	parallel-perf-output/
	├── time-range-0
	│   ├── cmd.txt
	│   └── out.txt
	├── time-range-1
	│   ├── cmd.txt
	│   └── out.txt
	├── time-range-2
	│   ├── cmd.txt
	│   └── out.txt
	└── time-range-3
	    ├── cmd.txt
	    └── out.txt
	$ find parallel-perf-output -name cmd.txt | sort | xargs grep -H .
	parallel-perf-output/time-range-0/cmd.txt:perf script --time=,9466.504461499 --ns
	parallel-perf-output/time-range-1/cmd.txt:perf script --time=9466.504461500,9469.005396999 --ns
	parallel-perf-output/time-range-2/cmd.txt:perf script --time=9469.005397000,9471.506332499 --ns
	parallel-perf-output/time-range-3/cmd.txt:perf script --time=9471.506332500, --ns

Any perf script command can be used, including the use of perf script options
--dlfilter and --script, so that the benefit of running parallel jobs
naturally extends to them also.

If option --pipe-to is used, standard output is first piped through that
command. Beware, if the command fails (e.g. grep with no matches), it will be
considered a fatal error.

Final standard output is redirected to files named out.txt in separate
subdirectories under the output directory. Similarly, standard error is
written to files named err.txt. In addition, files named cmd.txt contain the
corresponding perf script command. After processing, err.txt files are removed
if they are empty.

If any job exits with a non-zero exit code, then all jobs are killed and no
more are started. A message is printed if any job results in a non-empty
err.txt file.

There is a separate output subdirectory for each time range. If the --per-cpu
option is used, these are further grouped under cpu-n subdirectories, e.g.

	$ parallel-perf.py --per-cpu --nr=2 -- perf script --ns --cpu=0,1
	All jobs finished successfully
	$ tree parallel-perf-output
	parallel-perf-output/
	├── cpu-0
	│   ├── time-range-0
	│   │   ├── cmd.txt
	│   │   └── out.txt
	│   └── time-range-1
	│       ├── cmd.txt
	│       └── out.txt
	└── cpu-1
	    ├── time-range-0
	    │   ├── cmd.txt
	    │   └── out.txt
	    └── time-range-1
	        ├── cmd.txt
	        └── out.txt
	$ find parallel-perf-output -name cmd.txt | sort | xargs grep -H .
	parallel-perf-output/cpu-0/time-range-0/cmd.txt:perf script --cpu=0 --time=,9469.005396999 --ns
	parallel-perf-output/cpu-0/time-range-1/cmd.txt:perf script --cpu=0 --time=9469.005397000, --ns
	parallel-perf-output/cpu-1/time-range-0/cmd.txt:perf script --cpu=1 --time=,9469.005396999 --ns
	parallel-perf-output/cpu-1/time-range-1/cmd.txt:perf script --cpu=1 --time=9469.005397000, --ns

Subdivisions of time range, and cpus if the --per-cpu option is used, are
expressed by the --time and --cpu perf script options respectively. If the
supplied perf script command has a --time option, then that time range is
subdivided, otherwise the time range given by 'time of first sample' to
'time of last sample' is used (refer perf script --header-only). Similarly, the
supplied perf script command may provide a --cpu option, and only those CPUs
will be processed.

To prevent time intervals becoming too small, the --min-interval option can
be used.

Note there is special handling for processing Intel PT traces. If an interval is
not specified and the perf record command contained the intel_pt event, then the
time range will be subdivided in order to produce subdivisions that contain
approximately the same amount of trace data. That is accomplished by counting
double-quick (--itrace=qqi) samples, and choosing time ranges that encompass
approximately the same number of samples. In that case, time ranges may not be
the same for each CPU processed. For Intel PT, --per-cpu is the default, but
that can be overridden by --no-per-cpu. Note, for Intel PT, double-quick
decoding produces 1 sample for each PSB synchronization packet, which in turn
come after a certain number of bytes output, determined by psb_period (refer
perf Intel PT documentation). The minimum number of double-quick samples that
will define a time range can be set by the --min_size option, which defaults to
64.
""")
	ap.add_argument("-o", "--output-dir", default="parallel-perf-output", help="output directory (default 'parallel-perf-output')")
	ap.add_argument("-j", "--jobs", type=int, default=0, help="maximum number of jobs to run in parallel at one time (default is the number of CPUs)")
	ap.add_argument("-n", "--nr", type=int, default=0, help="number of time subdivisions (default is the number of jobs)")
	ap.add_argument("-i", "--interval", type=float, default=0, help="subdivide the time range using this time interval (in seconds e.g. 0.1 for a tenth of a second)")
	ap.add_argument("-c", "--per-cpu", action="store_true", help="process data for each CPU in parallel")
	ap.add_argument("-m", "--min-interval", type=float, default=glb_min_interval, help=f"minimum interval (default {glb_min_interval} seconds)")
	ap.add_argument("-p", "--pipe-to", help="command to pipe output to (optional)")
	ap.add_argument("-N", "--no-per-cpu", action="store_true", help="do not process data for each CPU in parallel")
	ap.add_argument("-b", "--min_size", type=int, default=glb_min_samples, help="minimum data size (for Intel PT in PSBs)")
	ap.add_argument("-D", "--dry-run", action="store_true", help="do not run any jobs, just show the perf script commands")
	ap.add_argument("-q", "--quiet", action="store_true", help="do not print any messages except errors")
	ap.add_argument("-v", "--verbose", action="store_true", help="print more messages")
	ap.add_argument("-d", "--debug", action="store_true", help="print debugging messages")
	cmd_line = list(args)
	try:
		split_pos = cmd_line.index("--")
		cmd = cmd_line[split_pos + 1:]
		args = cmd_line[:split_pos]
	except:
		cmd = None
		args = cmd_line
	a = ap.parse_args(args=args[1:])
	a.cmd = cmd
	a.verbosity = Verbosity(a.quiet, a.verbose, a.debug)
	try:
		if a.cmd == None:
			if len(args) <= 1:
				ap.print_help()
				return True
			raise Exception("Command line must contain '--' before perf command")
		return RunParallelPerf(a)
	except Exception as e:
		print("Fatal error: ", str(e))
		if a.debug:
			raise
		return False

if __name__ == "__main__":
	if not Main(sys.argv):
		sys.exit(1)
